;; Lispscript

;; A reliable messaging channel. Each Channel object represents one
;; end of a bidirectional message pipe. User code can use the
;; SEND-MESSAGE method on the Channel to send messages and receives
;; messages via the ON-MESSAGE callback.

;; The Channel doesn't need to know anything about the addressing of
;; the messages it sends and receives--it is simply given opaque
;; messages that it then arranges to send to the Channel on the other
;; side of the wire, once and only once, using an unreliable
;; transport. It sends two kinds of messages, SEND messages which
;; actually contain messages given to it to be sent, and ACK messages
;; which it uses to communicate to the Channel on the other side which
;; SEND messages it has, in fact, received.

;; The Channel needs to keep a queue of outgoing messages because it
;; can't necessarily send messages right away: if it has outstanding
;; unacked messages, it needs to hold on to new outgoing messages
;; until the acks come in. So a call to SEND-MESSAGE from the client
;; code may initially result only in the Channel queuing the message.
;; Eventually, however, the Channel will pass the message along to an
;; underlying transport mechanism which will attempt to covney the
;; message to other side.

;; Channel generates send and ack messages and hands them to the
;; http-transport immediately where they are held to be sent at the
;; next available opportunity.

;; The channel wraps the messages passed to it via SEND in an XML
;; envelope that contains a sequence number and then passes the
;; envelope to the HTTP-TRANSPORT object which actually sends the
;; messages. The channel also keeps 


;; Each time we send a message we enqueue it and then send it
;; asynchronously using an XMLHttpRequest object. We keep track of
;; messages actually sent, acknowledged, and received using three
;; counters. This allows us to detect gaps and dups: each message is
;; sent with a sequence number which is acked by the recipient. When
;; the sender recieves an ack, if it is the expected ack it removes
;; the message from the queue and sends the next message. Otherwise it
;; resends the message on the top of the queue, tagging it with the
;; same sequence number.




(defun message-dispatcher (channel message)
  (while (= (@ message node-type) 3)
    (set message (@ message next-sibling)))
  (let* ((message-type (@ message nodeName))
	 (handler (ref (@ this message-handlers) message-type)))
    (if handler
	(handler (@ message firstChild))
	(debug "No handler for message-type " message-type))))

(defun Channel (transport id)
  (set (@ this transport) transport)
  (.set-channel transport this)
  (set (@ this id) (or id (.randomID Channel)))
  (set (@ this sent) 0)
  (set (@ this acked) 0)
  (set (@ this received) 0)
  (set (@ this queue) (new Queue))
  (set (@ this send-timeout-millis) (* 10 1000))
  (set (@ this message-handlers) (new Object))
  (set (@ this on-message) message-dispatcher)
  (set (@ this logger) (lambda (text)))
  this)

(defmethod (Channel register-message-handler) (name fn)
  (set (ref (@ this message-handlers) name) fn))

(defmethod (Channel log) (text)
  ((@ this logger) text))

(define-class-method (Channel randomID) () 
  (+ (new Number (new Date)) "-" (.round Math (* (.random Math) (.pow Math 2 57)))))

;; This is the public API for sending messages. The payload should be
;; an XML DOM object representing the message to be sent. It will be
;; wrapped in a SEND message with an appropriate sequence number by
;; the Channel.

(defmethod (Channel send) (payload)
  (.log this (+ "[send] " payload))
  (let* ((sequence (.next-sequence-number this))
	 (message (make-send-xml sequence payload)))
    (.enqueue (@ this queue) message)
    (when (= (@ this sent) (@ this acked))
      (++ (@ this sent))
      (.send-send this sequence))))

;; This method is called by the transport for each incoming message.
(defmethod (Channel receive) (message-xml)
  (.log this (+ "[receive] " message-xml))
  (let ((type (@ message-xml node-name))
	(sequence (parse-int (.get-attribute message-xml "sequence"))))
    (cond
      ((or (string= type "SEND") (string= type "send"))
       (.normalize message-xml)
       (.receive-send this sequence (@ message-xml firstChild)))
      ((or (string= type "ACK") (string= type "ack"))
       (.receive-ack this sequence))
      (true
       (alert (+ "Unknown message type: " type))))))



;; When client code asks us to send a message we need to assign it a
;; sequence number. The next sequence number is one greater than the
;; last acked sequence plus the number of unacked messages, i.e. acked
;; plus the queue depth plus one.
(defmethod (Channel next-sequence-number) ()
  (1+ (+ (@ this acked) (.depth (@ this queue)))))

(defmethod (Channel send-send) (sequence)
  (.log this (+ "[send-send] " sequence))
  (.send (@ this transport) (.peek (@ this queue)))
  (.set-ack-timeout this sequence))

(defmethod (Channel send-ack) (sequence)
  (.log this (+ "[send-ack] " sequence))
  (.send (@ this transport) (make-ack-xml sequence)))

(defmethod (Channel set-ack-timeout) (sequence)
  ;; set a timeout that will fire in N seconds and resend the message
  ;; if it hasn't been acked.
  (let ((channel this))
    (.log channel (+ "[set-ack-timeout] will fire in " (@ this send-timeout-millis) " milliseconds."))
    (.set-timeout 
     window
     (lambda () 
       (cond
	 ((>= (@ channel acked) sequence)
	  (.log channel (+ "[ack-timeout] ack recieved.")))
	 (true
	  (.log channel (+ "[ack-timeout] ack not received, resending."))
	  (.send-send channel sequence))))
     (@ this send-timeout-millis))))

(defmethod (Channel receive-send) (sequence payload)
  (let ((expected (1+ (@ this received))))
    (cond
      ((= sequence expected)
       (.log this (+ "[receive-send@" (@ this sent) ":" (@ this acked) "] seq: " sequence ". Expected message, dispatching."))
       ((@ this on-message) this payload)
       (++ (@ this received))
       (.send-ack this (@ this received)))
      ((> sequence expected)
       (.log this (+ "[receive-send@" (@ this sent) ":" (@ this acked) "] seq: " sequence ". Gap, re-acking " (@ this received)))
       (.send-ack this (@ this received)))
      ((< sequence expected)
       (.log this (+ "[receive-send@" (@ this sent) ":" (@ this acked) "] seq: " sequence ". Dup, ignoring."))))))

(defmethod (Channel receive-ack) (ack)
  (cond
    ((and (= ack (@ this sent)) (> ack (@ this acked))) ;; message received.
     (.log this (+ "[receive-ack@" (@ this sent) ":" (@ this acked) "] ack: " ack ". Expected ack, removing message from queue."))
     (++ (@ this acked))
     (.dequeue (@ this queue))
     (when (not (.empty-p (@ this queue)))
       (++ (@ this sent))
       (.send-send this (@ this sent))))
    (true
     (.log this (+ "[receive-ack@" (@ this sent) ":" (@ this acked) "] ack: " ack ". Supurious ack, ignoring.")))))

(defmethod (Channel toString) ()
  (+ (@ this id) " [" (@ this sent) "/" (@ this acked) "/" (@ this received) "]; depth: " (.depth (@ this queue))))

(defun make-send-xml (sequence payload)
  (let ((message (.create-document (@ document implementation) "" "SEND" null)))
    (.set-attribute (@ message document-element) "SEQUENCE" (+ "" sequence))
    (.append-child 
     (@ message document-element)
     (.importNode message (@ payload document-element) true))
    message))

(defun make-ack-xml (sequence)
  (let ((message (.create-document (@ document implementation) "" "ACK" null)))
    (.set-attribute (@ message document-element) "SEQUENCE" (+ "" sequence))
    message))

(defun make-simple-message-for-app (app value)
  (let ((message (.create-document (@ document implementation) "" app null)))
    (.append-child (@ message document-element) (.create-text-node message value))
    message))

(defun make-message-for-app (app value)
  (let ((message (.create-document (@ document implementation) "" app null)))
    (.append-child 
     (@ message document-element)
     (.importNode message (@ value document-element) true))
    message))
