;;
;; Copyright (c) 2005, Gigamonkeys Consulting All rights reserved.
;;

(in-package :com.gigamonkeys.markup)

(defvar *notes*)
(defvar *note-counter*)

(defparameter *tag-translations* `((:example     . :pre)
				   (:sidebarhead . (:div :class "sidebarhead"))
				   (:sidebar     . (:div :class "sidebar"))
				   (:note-ref    . :sup)
				   (:bullets     . :ul)
				   (:numbered    . :ol)
				   (:note        . htmlize-note)
				   (:item        . htmlize-li)
				   (:url         . htmlize-url)
 				   (:img         . htmlize-img)
				   (:link        . htmlize-link)))


(defun htmlize-note (tag body)
  (declare (ignore tag))
  (push `((:div :class "note") ,@body) *notes*)
  `(:sup ,(incf *note-counter*)))

(defun htmlize-li (tag body)
  (declare (ignore tag))
  (if (single-paragraph-p body)
      `(:li ,@(cdar (translate-tags body)))
      `(:li ,@(translate-tags body))))

(defun single-paragraph-p (body)
  (and (= 1 (length body))
       (consp (car body))
       (member (caar body) *paragraph-tags*)))

(defun htmlize-url (tag body)
  (declare (ignore tag))
  `(:a :href ,@body ,@body))

(defun htmlize-img (tag body)
  (declare (ignore tag))
  `(:img :src ,@body))

(defun htmlize-link (tag body)
  (declare (ignore tag))
  (destructuring-bind ((tag1 &rest part1) (tag2 &rest part2)) body
    (let ((plist (list tag1 part1 tag2 part2)))
      (let ((href (getf plist :href))
	    (text (getf plist :text))
	    (name (getf plist :name)))
	(cond 
	  (href `(:a :href ,@href ,@text))
	  (name `(:a :name ,@name ,@text)))))))

(defmethod render-as ((type (eql :html)) sexp file)
  (with-html-to-file (file)
    (let ((*style-sheet-name* (or *style-sheet-name* (format nil "~a.css" (pathname-name file)))))
      (emit-html (make-foo-html (first (encode-characters sexp)))))))

(defun make-foo-html (sexp)
  (let ((*notes* ())
	(*note-counter* 0)
	(title (find-first-h1 sexp)))
    `(:html
       (:head 
	,@(when *custom-headers* `(,*custom-headers*))
	,@(when title `((:title ,title)))
	(:link :rel "stylesheet" :type "text/css" :href ,*style-sheet-name*))
       (:body
	,@(translate-tags sexp)
	,@(when *notes*
		`((:hr)
		  ,@(loop for note in 
			 (let ((*tag-translations* 
				(cons '(:note . (:div :class "note"))
				      (remove :note *tag-translations* :key #'car))))
			   (translate-tags (reverse *notes*)))
		       for number from 1 
		       do (push `(:sup ,(format nil "~d. " number)) (cdadr note))
		       collect note)))))))

(defun find-first-h1 (parse)
  (labels ((walk (thing)
	     (when (consp thing)
	       (destructuring-bind (tag &rest body) thing
		 (if (eql tag :h1)
		   (return-from find-first-h1 (car body))
		   (dolist (child body)
		     (walk child)))))))
    (dolist (thing parse)
      (walk thing))))

(defun translate-tag (element)
  (if (atom element) 
      element
      (destructuring-bind (tag &rest body) element
	(let ((translation (cdr (assoc tag *tag-translations*))))
	  (etypecase translation
	    (null
	     (cons tag (translate-tags body)))
	    (keyword
	     (cons translation (translate-tags body)))
	    (cons
	     (cons translation (translate-tags body)))
	    ((or symbol function)
	     (multiple-value-bind (translation recursive-p)
		 (funcall translation tag body)
	       (if recursive-p
		   (translate-tag translation)
		   translation))))))))

(defun translate-tags (list)
  (mapcar #'translate-tag list))

(defun replace-characters (string)
  #+(or)(list string)
  (loop for start = 0 then (+ end (length "--"))
     for end = (search "--" string :start2 start)
     collect (subseq string start end)
     while end collect '(:character :mdash)))

(defun encode-characters (tree)
  (cond
    ((stringp tree) (replace-characters tree))
    ((consp tree) 
     (if (member (car tree) *verbatim-elements*)
	 (list tree)
	 (list (mapcan #'encode-characters tree))))
    (t (list tree))))
