;;
;; Copyright (c) 2005, Gigamonkeys Consulting All rights reserved.
;;

(in-package :com.gigamonkeys.markup)

(defparameter *page-width-inches* 8.5)
(defparameter *default-font-size* 11)

(defparameter *margins* '(72  72 72 72)) ; left top right bottom


(defvar *no-mdash* nil)
(defvar *significant-whitespace* nil)

(defconstant +mdash+ (code-char 151))
(defconstant +copyright+ (code-char 169))
(defconstant +bullet+ (code-char 127))

(defparameter *my-leading-ratio* 1.2)

(defparameter *character-styles* ())
(defparameter *paragraph-styles* ())

(defvar *list-type* :bullets)
(defvar *list-number*)

(defparameter *table-cell-padding* 3)
(defparameter *table-padding* 20)

(defun add-character-style (tag style)
  (if (assoc tag *character-styles*)
      (setf (cdr (assoc tag *character-styles*)) style)
      (setf *character-styles* (acons tag style *character-styles*))))

(defun get-character-style (tag)
  (let ((style (cdr (assoc tag *character-styles*))))
    (typecase style
      (null style)
      (keyword (get-character-style style))
      (cons style)
      (function (funcall style)))))

(defun add-paragraph-style (tag style)
  (if (assoc tag *paragraph-styles*)
      (setf (cdr (assoc tag *paragraph-styles*)) style)
      (setf *paragraph-styles* (acons tag style *paragraph-styles*))))

(defun get-paragraph-style (tag)
  (let ((style (cdr (assoc tag *paragraph-styles*))))
    (typecase style
      (null style)
      (keyword (get-paragraph-style style))
      (cons style)
      (function (funcall style)))))


(add-character-style :i '(:font "Times-Italic"))
(add-character-style :b #'(lambda () `(:font ,(bold-font-name (pdf:name typeset::*font*)))))
(add-character-style :title :i)
(add-character-style :code '(:font "Courier"))
(add-character-style :cl '(:font "Courier-Bold"))

(add-paragraph-style :h1 '(:h-align :left :font "Helvetica" :font-size 18))
(add-paragraph-style :h2 '(:h-align :left :font "Helvetica" :font-size 16 :top-margin 18))
(add-paragraph-style :h3 '(:h-align :left :font "Helvetica" :font-size 12 :top-margin 12))
(add-paragraph-style :p #'(lambda () `(:h-align :left :font "Times-Roman" :font-size ,*default-font-size* :top-margin 12)))
(add-paragraph-style :blockquote #'(lambda () `(:h-align :left :font "Times-Roman" :font-size ,(1- *default-font-size*) :left-margin 18 :top-margin 12)))
(add-paragraph-style :example #'(lambda () `(:h-align :left :font "Courier" :font-size ,(- *default-font-size* 4) :left-margin 18 :top-margin 12)))
(add-paragraph-style :sidebar #'(lambda () `(:h-align :left :font "Helvetica" :font-size ,(1- *default-font-size*) :left-margin 18 :top-margin 12)))
(add-paragraph-style :sidebarhead '(:h-align :left :font "Helvetica-Oblique" :font-size 10 :left-margin 18))
(add-paragraph-style :item #'(lambda () `(:h-align :left :font "Times-Roman" :font-size ,*default-font-size* :left-margin 24)))
(add-paragraph-style :tablecaption '(:h-align :left :font "Helvetica-Oblique" :font-size 9 :top-margin 3))
(add-paragraph-style :figurecaption '(:h-align :left :font "Helvetica-Oblique" :font-size 9 :top-margin 3))
(add-paragraph-style :td '(:h-align :left :font "Times-Roman" :font-size 10))

(defmacro with-dynamic-style (style &body body)
  (with-gensyms (new-style restore-style)
    `(let* ((,new-style (apply #'typeset::make-instance 'typeset::text-style ,style))
	    (,restore-style (typeset::make-restore-style ,new-style)))
      (typeset::add-box ,new-style)
      (typeset::use-style ,new-style)
      ,@(mapcar 'typeset::insert-stuff body)
      (typeset::add-box ,restore-style)
      (typeset::use-style ,restore-style))))

(defmacro dynamic-paragraph (style &body body)
  (once-only (style)
    (with-gensyms (top-margin bottom-margin first-line-indent 
			      new-style restore-style first-indent)
      `(let* ((,top-margin (getf ,style :top-margin 0))
	      (,bottom-margin (getf ,style :bottom-margin 0))
	      (,first-line-indent (getf ,style :first-line-indent 0))
	      (,new-style (apply #'typeset::make-instance 'typeset::text-style ,style))
	      (,restore-style (typeset::make-restore-style ,new-style))
	      (,first-indent ,first-line-indent))
	(typeset::add-box ,new-style)
	(typeset::use-style ,new-style)
	(typeset::add-box (make-instance 'typeset::v-spacing :dy ,top-margin))
	(unless (zerop ,first-indent)
	  (typeset::add-box (make-instance 'typeset::h-spacing :dx ,first-indent)))
	,@(mapcar 'typeset::insert-stuff body)
	(unless (eq (first (typeset::boxes-tail typeset::*content*)) :eol)
	  (typeset::add-box :eol))
	(typeset::add-box (make-instance 'typeset::v-spacing :dy ,bottom-margin))
	(typeset::add-box ,restore-style)
	(typeset::use-style ,restore-style)))))

(defun standard-header (page)
  (declare (ignore page))
  (if (= pdf:*page-number* 1)
      (compile-text ())
      (compile-text (:font "Times-Italic" :font-size 10) 
	(paragraph (:h-align :fill)
	  ;; Not clear whether these should
	  ;; be VERBATIM or PUT-STRING. Get
	  ;; weird spacing with PUT-STRING.
	  (verbatim (get-contextual-variable 'chapter-name))
	  :hfill
	  (verbatim (format nil "Page ~d of ~d" 
			    pdf:*page-number*
			    (find-ref-point-page-number :the-end)))))))

(defun standard-footer (page)
  (declare (ignore page))
  (compile-text (:font "Times-Italic" :font-size 10)          
    (paragraph (:h-align :fill)
      (verbatim
       (format nil "Copyright ~c ~a" +copyright+ *copyright*))
      :hfill
      (verbatim (date-string)))))

(defparameter *pdf-header* #'standard-header)
(defparameter *pdf-footer* #'standard-footer)

(defmethod render-as ((type (eql :pdf)) sexp file)
  (multiple-value-bind (body notes) (extract-notes sexp)
    (with-document ()
      (let ((content
             (compile-text ()
               (let ((*my-leading-ratio* 1.2))
                 (dolist (p body) (emit-pdf p)))
               (vspace 24)
               (hrule :dy .1)
               (vspace 6)
               (let ((*default-font-size* 10))
                 (loop for n in notes
		    for counter from 1
		    do (emit-note counter n)))
               (mark-ref-point :the-end))))
	(draw-pages content
		    :break :after
		    :size :Letter
		    :margins *margins*
		    :header-top 36
		    :footer-bottom 36
		    :header *pdf-header*
		    :footer *pdf-footer*))
      (pdf:write-document file))))







(defgeneric emit-pdf (thing)
  (:documentation "Emit part of a sexp-marked document as PDF."))

(defmethod emit-pdf ((thing string))
  (if (or *significant-whitespace* *no-mdash*)
    (verbatim thing)
    (fix-character-typography thing)))

(defmethod emit-pdf ((thing character))
  (emit-pdf (string thing)))

(defmethod emit-pdf ((thing cons))
  (emit-pdf-by-type (car thing) (cdr thing)))

(defgeneric emit-pdf-by-type (type children)
  (:documentation "Called by emit-pdf for tagged elements."))

(defmethod emit-pdf-by-type ((type t) children)
  (cond
    ((get-character-style type)
     (with-dynamic-style (get-character-style type)
       (dolist (c children) (emit-pdf c))))
    ((get-paragraph-style type)
     (dynamic-paragraph (get-paragraph-style type)
       (dolist (c children) (emit-pdf c))))
    (t
     (error "No style for ~a" type))))

;;; Paragraph styles

(defmethod emit-pdf-by-type :before ((type (eql :h1)) children)
  ;;(typeset::fresh-page)
  (set-contextual-variable 'chapter-name (first children)))

(defmethod emit-pdf-by-type :around ((type (eql :p)) children)
  (declare (ignore children))
  (let ((typeset::*leading-ratio* *my-leading-ratio*))
    (call-next-method)))

(defmethod emit-pdf-by-type :around ((type (eql :example)) children)
  (declare (ignore children))
  (let ((*significant-whitespace* t))
    (call-next-method)))


(defmethod emit-pdf-by-type ((type (eql :bullets)) children)
  (vspace 12)
  (loop for c in children do (emit-pdf c)))

(defmethod emit-pdf-by-type ((type (eql :numbered)) children)
  (let ((*list-type* :numbered)
        (*list-number* 0))
    (vspace 12)
    (loop for (c next) on children
       for number from 1
       do (emit-pdf c)
       (when next (vspace 6)))))

(defmethod emit-pdf-by-type ((type (eql :item)) children)
  (cond
    ((= 1 (length children))
     (setf children (list* (case *list-type*
				   (:bullets
				    (format nil "~c " +bullet+))
				   (:numbered
				    (format nil "~d. " (incf *list-number*))))
			   (cdr (first children)))))
    (t
     (destructuring-bind ((first-tag &rest first-body) &rest rest) children
       
       
       (setf children (list* (list* first-tag
				     (case *list-type*
				       (:bullets
					(format nil "~c " +bullet+))
				       (:numbered
					(format nil "~d. " (incf *list-number*))))
				     first-body)
			      rest)))))
  (dynamic-paragraph (get-paragraph-style type)
    (dolist (c children) (emit-pdf c))))

;;; Tables

(defmethod emit-pdf-by-type ((type (eql :table)) children)
  (destructuring-bind (left top right bottom) *margins*
    (declare (ignore top bottom))
    (let* ((num-columns (length (cdr (first children))))
	   (width (- (* *page-width-inches* 72) left right (* num-columns (* 2 *table-cell-padding*)) (* 2 *table-padding*)))
	   (column-widths (make-list num-columns :initial-element (floor (/ width num-columns)))))
      (vspace 12)
      (table (:col-widths column-widths :background-color '(1 1 1) :cell-padding *table-cell-padding* :padding *table-padding* :splittable-p t)
	(dolist (c children)
	  (emit-pdf c))))))
  
(defmethod emit-pdf-by-type ((type (eql :tr)) children)
  (row () (dolist (c children) (emit-pdf c))))

(defmethod emit-pdf-by-type :around ((type (eql :td)) children)
  (declare (ignore children))
  (cell (:background-color '(1.0 1.0 1.0))
    (call-next-method)))

;;; Character styles

(defmethod emit-pdf-by-type :around ((type (eql :code)) children)
  (declare (ignorable children))
  (let ((*no-mdash* t)) (call-next-method)))

(defmethod emit-pdf-by-type ((type (eql :url)) children)
  (emit-pdf-by-type :code children))

(defmethod emit-pdf-by-type ((type (eql :link)) children)
  (emit-pdf (second children)))

(defmethod emit-pdf-by-type ((type (eql :note-ref)) children)
  (with-superscript ()
    (emit-pdf (princ-to-string (first children)))))

(defun emit-note (counter note)
  (destructuring-bind ((first-type &rest first-body) &rest rest) (rest note)
    (dolist (c `((,first-type ,(format nil "~d. " counter) ,@first-body) ,@rest))
      (emit-pdf c))))

(defun date-string (&optional (utc (get-universal-time)))
  (multiple-value-bind (sec min hour date month year day daylight-p zone)
      (decode-universal-time utc)
    (declare (ignorable sec min hour date month year day daylight-p zone))
    (format nil "~d ~[~;~
                      January~;~
                      February~;~
                      March~;~
                      April~;~
                      May~;~
                      June~;~
                      July~;~
                      August~;~
                      September~;~
                      October~;~
                      November~;~
                      December~;~] ~4d, ~d:~2,'0d ~:[am~;pm~]" date month year (1+ (mod (1- hour) 12)) min (>= hour 12))))

(defun bold-font-name (font-name)
  (cond
    ((string-equal font-name "Times-Roman") "Times-Bold")
    (t (format nil "~a-Bold" font-name))))

(defun fix-character-typography (string)
  (loop for start = 0 then (+ end 2)
     for end = (search "--" string :start2 start)
     do (put-string (subseq string start end))
     when end do (put-string (format nil "~c" +mdash+))
     while end))



(defun null-content (page)
  (declare (ignore page))
  (compile-text ()))